
#include "vconf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


vconf_t* new_vconf(
	void *(VHASH_CDECL *vconf_malloc) (size_t),
	void (VHASH_CDECL *vconf_free) (void *))
{
	vconf_t *v = 0;

	v = (vconf_t *) vconf_malloc(sizeof(vconf_t));

	v->malloc = vconf_malloc;
	v->free = vconf_free;

	v->filename=0;
 	v->options_hash = new_vhash(1, v->malloc, v->free);

 	return v;
}

static void delete_options_helper(
	void * option, 
	const char * option_name, 
	void * user_data)
{
	vconf_t *v = (vconf_t*)user_data;

	/*
	printf("%s %s %p\n", (char*)option_name, (char*)option, option);
	*/
	vhash_remove_string_key(v->options_hash, option_name);
	v->free(option);
}

void delete_vconf(
	vconf_t* v)
{
	if(v->filename)
		v->free((void*)v->filename);

	vhash_string_key_map_function(
			v->options_hash,
			delete_options_helper, v);

	delete_vhash(v->options_hash) ;

	v->free(v);
}


void vconf_clear_options(
	vconf_t* v)
{
	/*clear old options if any*/
	vhash_string_key_map_function(
			v->options_hash,
			delete_options_helper, v);
}

static void write_options_helper(
	void * option, 
	const char * option_name, 
	void * user_data)
{
	FILE *f = (FILE*)user_data;

	/*printf("%s \"%s\"\n", (char*)option_name, (char*)option);*/
	fprintf(f, "%s \"%s\"\n", (char*)option_name, (char*)option);
}
		
int vconf_write_file(
	vconf_t* v, 
	const char * filename)
{
	FILE * f = fopen(filename, "w");

	if(f == NULL) return 0;

	/*clear old options if any*/
	vhash_string_key_map_function(
			v->options_hash,
			write_options_helper, (void*)f);

	fclose(f);

	return 1;
}

int vconf_read_file(
	vconf_t* v, 
	const char * filename)
{

	int filename_length = (int) strlen(filename);

	if(v->filename)
		 v->free((void*)v->filename);

	v->filename = (const char*)v->malloc(filename_length +1);

	strcpy((char*)v->filename, filename);

	if(v->filename[0] != '\0'){

		FILE * file = 0;

		int read_again = 1;

		char * line_buffer = 0;
		unsigned long line_buffer_size = 0;

		int a_char;
		unsigned long n;
		unsigned long i;

		char * option_type;
		char * option;

		file = fopen(v->filename, "rb");

		if(!file) return 0;

		fseek(file, 0, SEEK_SET);

		/*TODO make bigger later*/
		line_buffer_size = 1;
		line_buffer = (char*) v->malloc(line_buffer_size);

READ_A_LINE:
		n=0;
		for(;;) {

			/*resize buffer as needed*/
			if(n == line_buffer_size) {
				char * new_line_buffer;

				new_line_buffer = (char*) v->malloc(line_buffer_size*2);
				memcpy(new_line_buffer, line_buffer, line_buffer_size);

				v->free(line_buffer);

				line_buffer = new_line_buffer;
				line_buffer_size = line_buffer_size*2;
			}
			
			a_char = fgetc(file);
			
			if(a_char == EOF) {
				read_again=0;
				line_buffer[n] = '\0';
				break;
			}


			if(a_char == 0x0D || a_char == 0x0A){
				if(n == 0) continue;
				line_buffer[n] = '\0';
				break;
			}

			if(a_char == '\t' || a_char == '\0') a_char = ' ';

			line_buffer[n] = (char) a_char;

			n++;
		}/*while*/
		
		i = 0;

		while(i<n && line_buffer[i] == ' ') i++;

		if(line_buffer[i] == '\0') goto PARSE_DONE;

		if(line_buffer[i] == ';' || line_buffer[i] == '#') goto PARSE_DONE;

		/*
		printf("%s\n",&line_buffer[i]);
		*/

		option_type = &line_buffer[i];

		while(i<n && (line_buffer[i] != ' ' && line_buffer[i] != '=')) i++;

		line_buffer[i] = '\0';
		/*default empty option*/
		option = &line_buffer[i];
		i++;

		/*once for space and once for = once more for space*/
		while(i<n && (line_buffer[i] == ' ' || line_buffer[i] == '=')) i++;
		while(i<n && (line_buffer[i] == ' ' || line_buffer[i] == '=')) i++;
		while(i<n && (line_buffer[i] == ' ' || line_buffer[i] == '=')) i++;

		if(line_buffer[i] == ';' || line_buffer[i] == '#') goto USE_PARSE;


		if(line_buffer[i] == '"') {
			/*quoted option*/
			i++;
			/*real option start*/
			option = &line_buffer[i];
			while(i<n && line_buffer[i] != '"') i++;
			line_buffer[i] = '\0';
			goto USE_PARSE;
		}

		if(line_buffer[i] == '\'') {
			/*quoted option*/
			i++;
			/*real option start*/
			option = &line_buffer[i];
			while(i<n && line_buffer[i] != '\'') i++;
			line_buffer[i] = '\0';
			goto USE_PARSE;
		}

		/*unquoted option*/
		/*real option start*/
		option = &line_buffer[i];
		while(i<n && (line_buffer[i] != ' ' && line_buffer[i] != '#' && line_buffer[i] != ';')) i++;
		line_buffer[i] = '\0';

USE_PARSE:
		{
			char * new_option = 0;
			char * old_option = 0;

			new_option = (char*) v->malloc(strlen(option)+1);

			strcpy(new_option, option);


			old_option = (char*) vhash_remove_string_key(v->options_hash, option_type);


			vhash_insert_string_key(v->options_hash, option_type, new_option);

			/*
			printf("%s %s %s %p %p\n", option_type, new_option, (char*)old_option ,new_option, old_option);
			*/

			if(old_option) 
				v->free(old_option);
		}


PARSE_DONE:
		if(read_again)
			goto READ_A_LINE;

		fclose(file);

		if(line_buffer) free(line_buffer);

	}

	return 1;
}

const char * vconf_get_filename(
	vconf_t* v)
{
	return v->filename;
}

vhash_t * vconf_get_option_hash(
	vconf_t* v)
{
	return v->options_hash;
}


const char * vconf_get_option(
	vconf_t* v, 
	const char * option)
{
	return (const char *) vhash_lookup_string_key(v->options_hash, option);
}

void vconf_set_option(
	vconf_t* v, 
	const char * option_type,
	const char * option)
{
	char * old_option = 0;
	char * new_option = 0;

	new_option = (char*) v->malloc(strlen(option)+1);

	strcpy(new_option, option);

	old_option = (char*) vhash_remove_string_key(v->options_hash, option_type);

	vhash_insert_string_key(v->options_hash, option_type, new_option);

	/*
	printf("%s %s %s %p %p\n", option_type, new_option, (char*)old_option ,new_option, old_option);
	*/

	if(old_option) 
		v->free(old_option);

}








