// DWFCrawlHandler.h : Declaration of the CDWFCrawlHandler

#pragma once
#include "resource.h"       // main symbols

#include "DWFGoogleDesktopCrawl.h"
#include "GoogleDesktopSearchAPI.h"


#include "dwfcore/String.h"
#include "dwf/whiptk/fileext.h"
#include "dwf/package/reader/ObjectDefinitionReader.h"
using namespace DWFCore;
using namespace DWFToolkit;



// CDWFCrawlHandler

class ATL_NO_VTABLE CDWFCrawlHandler : 
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CDWFCrawlHandler, &CLSID_DWFCrawlHandler>,
	public ISupportErrorInfo,
	public IDispatchImpl<IDWFCrawlHandler, &IID_IDWFCrawlHandler, &LIBID_DWFGoogleDesktopCrawlLib, /*wMajor =*/ 1, /*wMinor =*/ 0>
{
public:
	CDWFCrawlHandler()
	{
	}

DECLARE_REGISTRY_RESOURCEID(IDR_DWFCRAWLHANDLER)


BEGIN_COM_MAP(CDWFCrawlHandler)
	COM_INTERFACE_ENTRY(IDWFCrawlHandler)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(ISupportErrorInfo)
END_COM_MAP()

// ISupportsErrorInfo
	STDMETHOD(InterfaceSupportsErrorInfo)(REFIID riid);

	DECLARE_PROTECT_FINAL_CONSTRUCT()

	HRESULT FinalConstruct()
	{
		return S_OK;
	}
	
	void FinalRelease() 
	{
	}

public:

	STDMETHOD(HandleFile)( BSTR bstrFullPath, IDispatch* pGoogleEventFactory );

private:

	HRESULT _postContent( const DWFString& zContent, const void* pThumbnailBuffer = NULL, size_t nThumbnailBytes = 0, const DWFString& zThumbnailMIME = L"" );

private:

	CComBSTR									_bstrFullPath;
	CComQIPtr<IGoogleDesktopSearchEventFactory>	_spGoogleEventFactory;

private:

	//
	// customized reader for object definition documents
	// this object tells the parser only to create and provide
	// us with DWFProperty objects
	//
	class _ObjectDefinitionCrawler : public DWFObjectDefinitionReader
	{
	public:

		_ObjectDefinitionCrawler();
		virtual ~_ObjectDefinitionCrawler();

		const DWFString& content();

		void provideProperties( const DWFString&    zID,
                                vector<DWFString>*  pPropertyRefs,
                                DWFProperty::tList* pPropertyList )
        throw();

	private:

		DWFString _zContent;

	};

	//
	// cusomtized reader for w2d graphic streams
	//
	class _W2DCrawler : public WT_FileExt
	{
	public:

		_W2DCrawler( DWFInputStream& rW2DStream );
		virtual ~_W2DCrawler();

		WT_Result crawl();
		const DWFString& content();
		void addContent( const DWFString& zContent );

		WT_Result process_stream_close (void);
		WT_Result process_stream_end_seek ();
		WT_Result process_stream_open (void);
		WT_Result process_stream_read (int desired_bytes, int& bytes_read, void* buffer);
		WT_Result process_stream_seek (int distance, int& amount_seeked);
		WT_Result process_stream_tell (unsigned long *current_file_pointer_position);
		WT_Result process_stream_write(int size, void const* buffer);

	private:

		static WT_Result _handleText( WT_Text& rText, WT_File& rFile );
		static WT_Result _handleLayer( WT_Layer& rLayer, WT_File& rFile );
		static WT_Result _handleNamedView( WT_Named_View& rView, WT_File& rFile );
		
		static WT_Result _handleAuthor( WT_Author& rInformational, WT_File& rFile );
		static WT_Result _handleComments( WT_Comments& rInformational, WT_File& rFile );
		static WT_Result _handleCopyright( WT_Copyright& rInformational, WT_File& rFile );
		static WT_Result _handleCreator( WT_Creator& rInformational, WT_File& rFile );
		static WT_Result _handleDescription( WT_Description& rInformational, WT_File& rFile );
		static WT_Result _handleKeywords( WT_Keywords& rInformational, WT_File& rFile );
		static WT_Result _handleTitle( WT_Title& rInformational, WT_File& rFile );
		static WT_Result _handleSubject( WT_Subject& rInformational, WT_File& rFile );
		static WT_Result _handleSourceFilename( WT_Source_Filename& rInformational, WT_File& rFile );

	private:

		size_t			_nBytesAvailable;
		DWFString		_zContent;
		DWFInputStream& _rW2DStream;
	};

};

OBJECT_ENTRY_AUTO(__uuidof(DWFCrawlHandler), CDWFCrawlHandler)
